from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING

import networkx
from angr.utils.graph import GraphUtils

from .edge import Edge, EdgeSort

if TYPE_CHECKING:
    from collections.abc import Sequence


class EdgeRouter:
    def __init__(self, graph, col_map, row_map, node_locs, max_col, max_row) -> None:
        """
        :param networkx.DiGraph graph:  The graph to route edges on.
        """
        self._graph = graph

        self._rows = row_map
        self._cols = col_map
        self._node_locations = node_locs
        self._max_row = max_row
        self._max_col = max_col

        # Mark whether a spot (col, row) is valid or not for putting edges.
        self._edge_valid = []
        # A map between spots (col, row) and all vertical edges at that spot.
        self.vertical_edges = []
        # A map between spots (col, row) and all horizontal edges at that spot.
        self.horizontal_edges = []

        self._in_edges = defaultdict(list)
        self._out_edges = defaultdict(list)

        self.edges = self._route_edges()

    def _route_edges(self):
        """
        Route edges between nodes.

        :return:                       None
        """

        self._prepare_edge_routing()

        edges = []

        for src, dst, data in self._graph.edges(data=True):
            sort = None
            if data.get("type", None) == "exception":
                sort = EdgeSort.EXCEPTION_EDGE
            edge = self._route_edge(src, dst, sort)
            edges.append(edge)

        self._set_in_edge_indices()
        self._set_out_edge_indices()

        return edges

    def _route_edge(self, src, dst, edge_sort=None):
        """
        Find a series of grids to route an edge from the source node to the destination node.

        :param src:         The beginning of the edge to route.
        :param dst:         The end of the edge to route.
        :return:            None
        """

        MOVE_LEFT = 0
        MOVE_RIGHT = 2

        # build the edge
        edge = Edge(src, dst, sort=edge_sort)

        start_col, start_row = self._node_locations[src]
        end_col, end_row = self._node_locations[dst]

        # start from the middle of the block
        start_col += 1
        end_col += 1
        # start from the next row
        start_row += 1

        start_idx = self._assign_edge_to(edge, "vertical", start_col, start_row, 0)
        edge.add_point(start_col, start_row, start_idx)

        if start_row < end_row:
            min_row, max_row = start_row, end_row
        else:
            max_row, min_row = start_row, end_row

        # find a vertical column to route the edge to the target node
        col = start_col
        if self._edge_available(col, min_row, max_row):
            pass
        else:
            offset = 1
            while True:
                if self._edge_available(col + offset, min_row, max_row):
                    col = col + offset
                    break
                if self._edge_available(col - offset, min_row, max_row):
                    col = col - offset
                    break
                offset += 1

        if col != start_col:
            # generate a line to move to the target column

            if start_col < col:
                min_col, max_col = start_col, col
                move = MOVE_RIGHT
            else:
                max_col, min_col = start_col, col
                move = MOVE_LEFT

            idx = self._assign_edge_to(edge, "horizontal", min_col, start_row, max_col - min_col)
            edge.add_point(col, start_row, idx)
            edge.add_move(move)
        else:
            # there will be a horizontal edge even when the beginning column and the target column are the same, since
            # the two blocks may not be aligned.
            _ = self._assign_edge_to(edge, "horizontal", start_col, start_row, 1)
            # however, we do not need to add the point to the edge

        if start_row != end_row:
            # generate a line to move to the target row
            idx = self._assign_edge_to(edge, "vertical", col, min(start_row + 1, end_row), abs(end_row - start_row) + 1)
            edge.add_point(col, end_row, idx)

        if col != end_col:
            # generate a line to move to the target column
            if col < end_col:
                min_col, max_col = col, end_col
                move = MOVE_RIGHT
            else:
                max_col, min_col = col, end_col
                move = MOVE_LEFT
            idx = self._assign_edge_to(edge, "horizontal", min_col, end_row, max_col - min_col)
            edge.add_point(end_col, end_row, idx)
            edge.add_move(move)

            # move downwards
            # in a new grid, we need a new edge index
            idx = self._assign_edge_to(edge, "vertical", end_col, end_row, 0)
            edge.add_point(end_col, end_row, idx)

        self._add_edge(edge)
        # print(src, dst, edge.points)

        return edge

    def _prepare_edge_routing(self) -> None:
        """
        Create and initialize necessary data structions for edge routing.

        :return: None
        """

        self._edge_valid = []
        for _col in range(self._max_col + 2):
            self._edge_valid.append([True] * (self._max_row + 1))
        for col, row in self._node_locations.values():
            # edges should not overlap with existing nodes
            self._edge_valid[col][row] = False
            self._edge_valid[col + 1][row] = False

        self.vertical_edges = []
        self.horizontal_edges = []

        for _col in range(self._max_col + 2):
            v_edges = []
            h_edges = []
            for _row in range(self._max_row + 3):
                v_edges.append({})
                h_edges.append({})
            self.vertical_edges.append(v_edges)
            self.horizontal_edges.append(h_edges)

    def _assign_edge_to(self, edge, sort, col, row, blocks, index=None):
        if sort == "vertical":
            d = self.vertical_edges
        elif sort == "horizontal":
            d = self.horizontal_edges
        else:
            raise ValueError(f'_assign_edge_to(): Unsupported edge sort "{sort}".')

        if sort == "vertical":
            if index is None:
                index = self._find_vertical_available_edge_index(col, row, row + blocks)
            for r in range(row, row + blocks + 1):
                d[col][r][index] = edge

        else:  # sort == 'horizontal'
            if index is None:
                index = self._find_horizontal_available_edge_index(col, col + blocks, row)
            for col_ in range(col, col + blocks + 1):
                d[col_][row][index] = edge

        return index

    def _edge_available(self, col, start_row, end_row):
        return all(self._edge_valid[col][i] for i in range(start_row, end_row))

    def _first_unused_index(self, indices):
        # find the first unused index
        last_i = None
        for i in sorted(indices):
            if last_i is None or i == last_i + 1:
                last_i = i
            else:
                # we found a gap
                return last_i + 1

        return 0 if last_i is None else last_i + 1

    def _find_vertical_available_edge_index(self, col, start_row, end_row):
        # collect all used indices
        indices = set()

        for row in range(start_row, end_row + 1):
            if self.vertical_edges[col][row]:
                indices.update(self.vertical_edges[col][row].keys())

        return self._first_unused_index(indices)

    def _find_horizontal_available_edge_index(self, start_col, end_col, row):
        # collect all used indices
        indices = set()

        for col in range(start_col, end_col + 1):
            if self.horizontal_edges[col][row]:
                indices.update(self.horizontal_edges[col][row].keys())

        return self._first_unused_index(indices)

    def _add_edge(self, edge) -> None:
        """
        Add an edge.

        :param Edge edge:   The Edge instance to add.
        :return:            None
        """

        self._out_edges[edge.src].append(edge)
        self._in_edges[edge.dst].append(edge)

    def _set_in_edge_indices(self) -> None:
        # assign in-edge indices
        for _, edges in self._in_edges.items():
            max_idx = None

            if len(edges) == 2:
                # sort by their last horizontal move
                edges = sorted(edges, key=lambda edge: edge.last_move, reverse=True)

            for idx, edge in enumerate(edges):
                edge.end_index = idx
                if max_idx is None or idx > max_idx:
                    max_idx = idx
            for edge in edges:
                edge.max_end_index = max_idx

    def _set_out_edge_indices(self) -> None:
        for _, edges in self._out_edges.items():
            max_idx = None
            if len(edges) == 2:
                edges = sorted(edges, key=lambda edge: edge.first_move)  # sort by their first horizontal move

            for idx, edge in enumerate(edges):
                edge.start_index = idx
                if max_idx is None or edge.start_index > max_idx:
                    max_idx = edge.start_index
            for edge in edges:
                edge.max_start_index = max_idx


class GraphLayouter:
    """
    Implements a pseudo layered graph layout (Sugiyama graph layout) algorithm.
    """

    def __init__(
        self,
        graph,
        node_sizes: Sequence[int],
        node_compare_key=None,
        node_sorter=None,
        x_margin: int = 10,
        y_margin: int = 5,
        row_margin: int = 16,
        col_margin: int = 16,
    ) -> None:
        self.graph = graph
        self._node_sizes = node_sizes
        self._node_compare_key = node_compare_key
        self._node_sorter = node_sorter

        if self._node_compare_key and self._node_sorter:
            raise RuntimeError("You cannot provide both node_compare_key and node_sorter.")

        self.x_margin = x_margin
        self.y_margin = y_margin
        self.row_margin = row_margin
        self.col_margin = col_margin

        self._cols = None
        self._rows = None
        self._max_col = None
        self._max_row = None
        self._locations = None

        self._vertical_edges = None
        self._horizontal_edges = None

        self._grid_max_vertical_id = {}
        self._grid_max_horizontal_id = {}
        self._row_to_nodes = {}
        self._row_heights = []
        self._col_widths = []
        self._grid_coordinates = {}

        self.edges: list[Edge] = []
        self.node_coordinates = {}

        self._layout()

    def _layout(self) -> None:
        self._initialize()

        # order the nodes
        ordered_nodes = GraphUtils.quasi_topological_sort_nodes(self.graph)

        # conver the graph to an acylic graph
        acyclic_graph = self._to_acyclic_graph(self.graph, ordered_nodes=ordered_nodes)

        # assign row and column to each node
        self._assign_grid_locations(self.graph, acyclic_graph, ordered_nodes=ordered_nodes)

        # edge routing
        edge_router = EdgeRouter(self.graph, self._cols, self._rows, self._locations, self._max_col, self._max_row)
        self.edges = edge_router.edges
        self._vertical_edges = edge_router.vertical_edges
        self._horizontal_edges = edge_router.horizontal_edges

        # determine the maximum index for each grid
        self._set_max_grid_edge_id()

        # determine row and column sizes
        self._make_grids()

        # calculate coordinates of nodes
        self._calculate_coordinates()

    def _initialize(self) -> None:
        self._cols = {}
        self._rows = {}
        self._locations = {}

    def _to_acyclic_graph(self, graph, ordered_nodes=None):
        """
        Convert a given DiGraph into an acyclic graph.

        :param networkx.DiGraph graph: The graph to convert.
        :param list ordered_nodes:     A list of nodes sorted in a topological order.
        :return:                       The converted acyclic graph.
        """

        if ordered_nodes is None:
            # take the quasi-topological order of the graph
            ordered_nodes = GraphUtils.quasi_topological_sort_nodes(graph)

        acyclic_graph = networkx.DiGraph()

        # add each node and its edge into the graph
        visited = set()
        for node in ordered_nodes:
            visited.add(node)
            acyclic_graph.add_node(node)
            for successor in graph.successors(node):
                if successor not in visited:
                    acyclic_graph.add_edge(node, successor)

        return acyclic_graph

    def _assign_grid_locations(self, graph, acyclic_graph, ordered_nodes=None) -> None:
        """
        Assign locations to each node in the graph in a bottom-up manner.

        :param networkx.DiGraph graph:          The original graph.
        :param networkx.DiGraph acyclic_graph:  The acyclic graph to work on.
        :param list ordered_nodes:      A list of nodes sorted in a topological order.
        :return:                        None
        """

        if ordered_nodes is None:
            # take the quasi-topological order of the graph
            ordered_nodes = GraphUtils.quasi_topological_sort_nodes(acyclic_graph)

        self._assign_rows(graph, acyclic_graph, ordered_nodes)
        self._assign_columns(acyclic_graph)

    def _assign_rows(self, graph, acyclic_graph, ordered_nodes) -> None:
        row_to_nodes = defaultdict(list)

        global_max_row = 0

        max_rows = {}

        """
        # assign min row ID top-down
        for node in ordered_nodes:
            if node not in min_rows:
                min_rows[node] = 0
            row = min_rows[node]
            for successor in acyclic_graph.successors_iter(node):
                if successor not in min_rows or min_rows[successor] > row + 1:
                    min_rows[successor] = row + 1
        """

        # assign max row ID using DFS
        for node in ordered_nodes:
            if node not in max_rows:
                max_rows[node] = 0
            row = max_rows[node]
            global_max_row = max(global_max_row, row)
            for successor in acyclic_graph.successors(node):
                if successor not in max_rows or max_rows[successor] < row + 1:
                    max_rows[successor] = row + 1
                    global_max_row = max(global_max_row, row + 1)

        self._max_row = global_max_row

        """
        for node in reversed(ordered_nodes):
            row = 0
            for successor in graph.successors_iter(node):
                succ_row = max_rows.get(successor, None)
                if succ_row is not None and succ_row + 1 > row:
                    row = succ_row + 1
            max_rows[node] = row
            global_max_row = max(global_max_row, row)

        self._max_row = global_max_row

        # invert row IDs
        for node in ordered_nodes:
            max_rows[node] = self._max_row - max_rows[node]
        """

        # determine row ID for each node
        for node in ordered_nodes:
            # we want to push each node as far up as possible, unless it is the return node
            row = max_rows[node]
            self._rows[node] = row
            row_to_nodes[row].append(node)

        for row in row_to_nodes:
            if self._node_compare_key is not None:
                row_to_nodes[row] = sorted(row_to_nodes[row], key=self._node_compare_key)
            elif self._node_sorter is not None:
                row_to_nodes[row] = self._node_sorter(row_to_nodes[row])
            else:
                # TODO: Use a custom comparator for displaying the CFG, too
                row_to_nodes[row] = sorted(row_to_nodes[row], key=lambda n_: n_.addr, reverse=True)

        self._row_to_nodes = row_to_nodes

    def _assign_columns(self, acyclic_graph) -> None:
        global_max_col = 0

        # First iteration: assign column ID bottom-up
        for row_idx in reversed(list(self._row_to_nodes.keys())):
            if self._node_compare_key is not None:
                row_nodes = sorted(self._row_to_nodes[row_idx], key=self._node_compare_key)
            elif self._node_sorter is not None:
                row_nodes = self._node_sorter(self._row_to_nodes[row_idx])
            else:
                row_nodes = sorted(self._row_to_nodes[row_idx], key=lambda n: n.addr)

            next_min_col, next_max_col = 1, 2

            for _i, node in enumerate(row_nodes):
                successors = acyclic_graph.successors(node)

                min_col, max_col = None, None

                for successor in successors:
                    if successor in self._cols:
                        succ_col = self._cols[successor]
                        if min_col is None or succ_col < min_col:
                            min_col = succ_col
                        if max_col is None or succ_col > max_col:
                            max_col = succ_col + 1

                if min_col is None and max_col is None:
                    min_col, max_col = next_min_col, next_max_col
                else:
                    if min_col < next_min_col:
                        min_col = next_min_col
                    if max_col < next_min_col:
                        max_col = next_min_col + 1

                # now assign a column ID to the current node
                col = (min_col + max_col) // 2
                self._cols[node] = col
                self._locations[node] = (col, row_idx)
                global_max_col = max(global_max_col, col)

                # update min_col and max_col for the next iteration
                next_min_col = max_col + 2 if min_col == max_col else max_col + 1
                next_max_col = next_min_col + 1

        # Second iteration: Adjust column IDs top-down
        for row_idx in self._row_to_nodes:
            row_nodes = self._row_to_nodes[row_idx]

            next_min_col, next_max_col = None, None

            for _i, node in enumerate(row_nodes):
                predecessors = list(acyclic_graph.predecessors(node))
                if len(predecessors) < 2:
                    # Not enough predecessors.
                    # update next_min_col and next_max_col
                    col = self._cols[node]
                    next_min_col = max(next_min_col if next_min_col is not None else 0, col + 2)
                    next_max_col = max(next_max_col if next_max_col is not None else 0, col + 3)
                    continue

                min_col, max_col = next_min_col, next_max_col

                for predecessor in predecessors:
                    if predecessor in self._cols:
                        pred_col = self._cols[predecessor]
                        if min_col is None or min_col > pred_col:
                            min_col = pred_col
                        if max_col is None or max_col < pred_col:
                            max_col = pred_col + 1

                # ideally, this node appears in between its predecessors
                col = (min_col + max_col) // 2
                overlap, col = self._detect_overlap(node, col, row_idx, min_col)

                # now assign a column ID to the current node
                self._cols[node] = col
                self._locations[node] = (col, row_idx)

                next_min_col = max_col + 1
                next_max_col = next_min_col + 1

                global_max_col = max(global_max_col, col)

        self._max_col = global_max_col + 1

    def _detect_overlap(self, node, ideal_col, row_idx, min_col):
        """
        Detect if any overlap will be caused if node in row_idx is placed at column col.

        :param node:            The node.
        :param int ideal_col:   The ideal column index.
        :param int row_idx:     The row that the node is placed at.
        :param int min_col:     The left-most acceptable column index for this node.
        :return:                (bool, int|None), where the first value is True if overlap is detected, False otherwise;
                                the second value is the suggested column.
        :rtype:                 tuple
        """

        overlap_detected = False
        suggested_col = min_col  # will only be used if overlap is detected

        # overlap detection
        for samerow_node in sorted(self._row_to_nodes[row_idx], key=lambda n: self._cols[n]):
            if samerow_node is node:
                continue
            samerow_node_col = self._cols[samerow_node]
            if samerow_node_col - 1 <= ideal_col <= samerow_node_col + 1:
                # collision detected :(
                overlap_detected = True
            if samerow_node_col - 1 <= suggested_col <= samerow_node_col + 1:
                # adjust our suggestion, which will be tested in the next iteration
                suggested_col = samerow_node_col + 2
            if overlap_detected and suggested_col < samerow_node_col - 1:
                # amazing, we got a suggestion working!
                # early termination of the loop
                break

        if overlap_detected:
            return True, suggested_col
        else:
            return False, ideal_col

    def _make_grids(self) -> None:
        """
        Determine the width of each column and the height of each row.

        :return: None
        """

        self._row_heights = [0] * (self._max_row + 2)
        self._col_widths = [0] * (self._max_col + 2)

        # update grid sizes based on nodes
        for node in self.graph.nodes():
            col, row = self._locations[node]

            width, height = self._node_sizes[node]

            if self._row_heights[row] < height:
                self._row_heights[row] = height

            if self._col_widths[col] < width // 2:
                self._col_widths[col] = width // 2
            if col + 1 < len(self._col_widths) and self._col_widths[col + 1] < width // 2:
                self._col_widths[col + 1] = width // 2

        # update grid sizes based on edges
        for col in range(self._max_col + 2):
            for row in range(self._max_row + 2):
                key = (col, row)
                if key in self._grid_max_vertical_id:
                    col_width = (self._grid_max_vertical_id[key] + 2) * self.x_margin
                    if self._col_widths[col] < col_width:
                        self._col_widths[col] = col_width
                if key in self._grid_max_horizontal_id:
                    row_height = (self._grid_max_horizontal_id[key] + 2) * self.y_margin
                    if self._row_heights[row] < row_height:
                        self._row_heights[row] = row_height

        # the left-most and the right-most column do not have any node assigned to it
        if self._col_widths[0] < 20:
            self._col_widths[0] = 20
        if self._col_widths[-1] < 20:
            self._col_widths[-1] = 20

    def _set_max_grid_edge_id(self) -> None:
        """
        For each grid, calculate the maximum edge index for both horizontal edges and vertical edges.

        :return: None
        """

        # horizontal edges
        for col, row_edges in enumerate(self._horizontal_edges):
            for row, edges in enumerate(row_edges):
                if not edges:
                    continue
                key = (col, row)
                self._grid_max_horizontal_id[key] = max(edges.keys())

        # vertical edges
        for col, row_edges in enumerate(self._vertical_edges):
            for row, edges in enumerate(row_edges):
                if not edges:
                    continue
                key = (col, row)
                self._grid_max_vertical_id[key] = max(edges.keys())

    def _calculate_coordinates(self):
        """
        Calculate coordinates for each grid, and then calculate coordinates for each node.

        :return: None
        """

        row_max_ids = {}
        for col, row in self._grid_max_horizontal_id:
            if row not in row_max_ids or self._grid_max_horizontal_id[(col, row)] > row_max_ids[row]:
                row_max_ids[row] = self._grid_max_horizontal_id[(col, row)]

        y = 0
        # calculate the top margin based on the number of horizontal edges above
        top_margin_height = self.row_margin * 2
        if 0 in row_max_ids:
            top_margin_height += self.y_margin * (row_max_ids[0] + 2)
        y += top_margin_height

        for row in range(-1, self._max_row + 2):
            x = 0

            for col in range(-1, self._max_col + 2):
                self._grid_coordinates[(col, row)] = (x, y)
                x += self._col_widths[col] + self.col_margin
            if self._row_heights[row] is None:
                self._row_heights[row] = 0

            # calculate the bottom margin based on the number of horizontal edges below
            bottom_margin_height = self.row_margin * 2
            if (row + 1) in row_max_ids:
                bottom_margin_height += self.y_margin * (row_max_ids[row + 1] + 2)
            y += self._row_heights[row] + bottom_margin_height

        # nodes
        for node in self.graph.nodes():
            col, row = self._locations[node]
            grid_x, grid_y = self._grid_coordinates[(col, row)]
            grid_a_width, grid_b_width = self._col_widths[col], self._col_widths[col + 1]
            grid_height = self._row_heights[row]
            node_width, node_height = self._node_sizes[node]

            self.node_coordinates[node] = (
                grid_x + ((grid_a_width + grid_b_width) // 2 - node_width // 2),
                grid_y + (grid_height // 2 - node_height // 2),
            )

        # edges
        for edge in self.edges:
            src_node_x, src_node_y = self.node_coordinates[edge.src]
            src_node_width, src_node_height = self._node_sizes[edge.src]

            dst_node_x, dst_node_y = self.node_coordinates[edge.dst]
            dst_node_width, dst_node_height = self._node_sizes[edge.dst]

            # dst_node_col, dst_node_row = self._locations[edge.dst]

            # start point
            _, _, start_x_index = edge.points[0]
            start_point_x_base = src_node_x + src_node_width // 2 - (self.x_margin * (edge.max_start_index + 1) // 2)
            start_point_x = self._indexed_x(start_point_x_base, start_x_index)
            start_point = (start_point_x, src_node_y + src_node_height)
            edge.add_coordinate(*start_point)

            prev_col, prev_row = self._locations[edge.src]
            prev_col += 1
            prev_row += 1
            x, y_base = start_point[0], start_point[1] + self.row_margin

            if len(edge.points) > 1:
                next_col, next_row, next_idx = edge.points[1]
                starting_col, starting_row = self._locations[edge.src]
                y_base = self._nointersecting_y(starting_row, starting_col, next_col, default=y_base) + self.row_margin
                y = self._indexed_y(y_base, next_idx)
            else:
                y = y_base

            # add a line that moves downwards from the exit
            edge.add_coordinate(x, y)

            # set previous x and y
            prev_x, prev_y = x, y

            # each point on the edge

            for point_id, (col, row, _) in enumerate(edge.points[1:-1]):
                if col == prev_col:
                    assert row != prev_row
                    # vertical
                    x = prev_x

                    base_y = self._grid_coordinates[(col, row - 1)][1] + self._row_heights[row - 1] + self.row_margin
                    if point_id + 1 == len(edge.points) - 2:
                        y = base_y
                    else:
                        next_col, next_row, next_idx = edge.points[point_id + 1 + 1]
                        y = self._indexed_y(base_y, next_idx)

                elif row == prev_row:
                    assert col != prev_col
                    # horizontal
                    if point_id + 1 == len(edge.points) - 2:
                        base_x = dst_node_x + dst_node_width // 2
                        x = self._indexed_x(base_x, edge.end_index)
                    else:
                        next_col, next_row, next_idx = edge.points[point_id + 1 + 1]
                        base_x = self._grid_coordinates[(col, row)][0]
                        x = self._indexed_x(base_x, next_idx)

                    y = prev_y

                else:
                    # the impossible branch
                    raise AssertionError

                edge.add_coordinate(x, y)

                # update prev_*
                prev_col, prev_row = col, row
                prev_x, prev_y = x, y

            # the last point, which is always at the top of the destination node
            base_x = dst_node_x + dst_node_width // 2 - (self.x_margin * (edge.max_end_index + 1) // 2)
            _, _, end_x_index = edge.points[-1]
            x = self._indexed_x(base_x, end_x_index)
            if x != prev_x:
                # add an extra coordinate to move horizontally
                edge.add_coordinate(x, prev_y)
            end_point = (x, dst_node_y - 6)
            edge.add_coordinate(*end_point)

    def _indexed_x(self, base_x, idx: int):
        return base_x + idx * self.x_margin

    def _indexed_y(self, base_y, idx: int):
        return base_y + idx * self.y_margin

    def _nointersecting_y(self, row, starting_col, ending_col, default=None):
        """
        Return the correct y coordinate for a point on an edge that will not lead to the following edge segment
        intersect with any node between `min_col` and `max_col`.

        :param int row:             Row of this point.
        :param int starting_col:    The starting column of the next edge segment.
        :param int ending_col:      The ending column of the next edge segment.
        :param float default:       The default y coordinate to use if we fail to determine a desired y coordinate.
        :return:                    The desired y coordinate.
        :rtype:                     float
        """

        max_y = None

        if starting_col < ending_col:
            min_col, max_col = starting_col, ending_col
        else:
            min_col, max_col = ending_col, starting_col

        for col in range(min_col, max_col + 1):
            key = (col, row)
            if key not in self._grid_coordinates:
                continue
            _, y = self._grid_coordinates[key]
            new_y = y + self._row_heights[row]
            if max_y is None or new_y > max_y:
                max_y = new_y

        return max_y if max_y is not None else default



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import networkx

from .job import InstanceJob

if TYPE_CHECKING:
    from angrmanagement.data.instance import Instance
    from angrmanagement.logic.jobmanager import JobContext


class DDGGenerationJob(InstanceJob):
    """A job that runs the VSA_DDG analysis for a function at a given address."""

    def __init__(self, instance: Instance, addr: int) -> None:
        super().__init__("DDG generation", instance, on_finish=self._finish)
        self._addr = addr

    def run(self, _: JobContext):
        ddg = self.instance.project.analyses.VSA_DDG(vfg=self.instance.vfgs[self._addr], start_addr=self._addr)
        return ddg, networkx.relabel_nodes(ddg.graph, lambda n: n.insn_addr)

    def _finish(self, result: Any) -> None:
        self.instance.ddgs[self._addr] = result

    def __repr__(self) -> str:
        return f"Generating VSA_DDG for function at {self._addr:#x}"



from __future__ import annotations

import functools
import logging
from typing import TYPE_CHECKING

from angrmanagement.data.jobs import SimgrExploreJob

from .debugger import Debugger

if TYPE_CHECKING:
    from angr import SimState

    from angrmanagement.ui.widgets.qsimulation_managers import QSimulationManagers
    from angrmanagement.ui.workspace import Workspace


_l = logging.getLogger(name=__name__)


class SimulationDebugger(Debugger):
    """
    Simulation debugger.
    """

    def __init__(self, sim_mgrs: QSimulationManagers, workspace: Workspace) -> None:
        super().__init__(workspace)
        self._sim_mgr_view: QSimulationManagers = sim_mgrs
        self._sim_mgr = sim_mgrs.simgr
        self._sim_mgr.am_subscribe(self._watch_simgr)
        self._sim_mgr_view.state.am_subscribe(self._watch_state)

    def __str__(self) -> str:
        if self._sim_mgr.am_none:
            return "No Simulation Manager"
        if self.simstate is None:
            return "Simulation (No active states)"
        else:
            pc = self.simstate.solver.eval(self.simstate.regs.pc)
            return f'Simulation @ {pc:x} ({len(self._sim_mgr.stashes["active"])} active)'

    def _watch_state(self, **_) -> None:
        self._on_state_change()

    def _watch_simgr(self, **_) -> None:
        self._on_state_change()

    def _on_state_change(self) -> None:
        """
        Common handler for state changes.
        """
        self.state_changed.am_event()

    @property
    def simstate(self) -> SimState:
        if not self._sim_mgr_view.state.am_none:
            return self._sim_mgr_view.state.am_obj
        elif not self._sim_mgr.am_none and len(self._sim_mgr.stashes["active"]) > 0:
            return self._sim_mgr.stashes["active"][0]
        else:
            return None

    @property
    def is_running(self) -> bool:
        return not self._sim_mgr.am_none

    @property
    def can_step_forward(self) -> bool:
        return not self._sim_mgr.am_none and self.is_halted and len(self._sim_mgr.stashes["active"]) > 0

    def step_forward(self, until_addr: int | None = None) -> None:
        if until_addr is not None:
            _l.warning("Step-until not implemented for SimulationDebugger")
        if self.can_step_forward:
            self._sim_mgr_view._on_step_clicked()

    @property
    def can_continue_forward(self) -> bool:
        return self.can_step_forward

    def continue_forward(self) -> None:
        if self.can_continue_forward:
            self._sim_mgr_view._on_explore_clicked()

    @property
    def _num_active_explore_jobs(self) -> int:
        return functools.reduce(lambda s, j: s + isinstance(j, SimgrExploreJob), self.workspace.job_manager.jobs, 0)

    @property
    def is_halted(self) -> bool:
        return self._num_active_explore_jobs == 0

    @property
    def can_halt(self) -> bool:
        return not self.is_halted

    def halt(self) -> None:
        for job in self.workspace.job_manager.jobs:
            if isinstance(job, SimgrExploreJob):
                self.workspace.job_manager.cancel_job(job)



from __future__ import annotations

import logging
import os
from typing import TYPE_CHECKING

import binharness
from PySide6.QtCore import Qt
from PySide6.QtGui import QIcon
from PySide6.QtWidgets import QComboBox, QDialog, QHBoxLayout, QLabel, QPushButton, QVBoxLayout

from angrmanagement.config import IMG_LOCATION

from .stream_view import StreamView

if TYPE_CHECKING:
    from angrmanagement.ui.workspace import Workspace

    from .bhinstance import BhInstance

log = logging.getLogger(name=__name__)


class RunTargetDialog(QDialog):
    """
    Dialog that allows the user to run a target with a specific environment and executor.
    """

    bh_instance: BhInstance

    def __init__(self, workspace: Workspace, bh_instance: BhInstance):
        super().__init__()
        self.setWindowFlags(Qt.WindowType.WindowTitleHint | Qt.WindowType.WindowCloseButtonHint)
        self.setWindowTitle("Run Target")
        # Icon
        angr_icon_location = os.path.join(IMG_LOCATION, "angr.png")
        self.setWindowIcon(QIcon(angr_icon_location))

        self.setMinimumWidth(400)

        self.bh_instance = bh_instance
        self.workspace = workspace
        self._init_widgets()

    def _init_widgets(self):
        # Select Environment
        self.env_combobox = QComboBox()
        self.env_combobox.addItems([env_id[0] for env_id in self.bh_instance.environments])

        # Select Executor
        self.executor_combobox = QComboBox()
        self.executor_combobox.addItems(["None"])

        # Run Button
        run_button = QPushButton("Run")
        run_button.clicked.connect(self._on_run_target)

        structure = QVBoxLayout()
        env_row = QHBoxLayout()
        env_row.addWidget(QLabel("Environment"))
        env_row.addWidget(self.env_combobox)
        exec_row = QHBoxLayout()
        exec_row.addWidget(QLabel("Executor"))
        exec_row.addWidget(self.executor_combobox)

        structure.addLayout(env_row)
        structure.addLayout(exec_row)
        structure.addWidget(run_button)

        layout = QHBoxLayout()
        layout.addLayout(structure)

        self.setLayout(layout)

    def _on_run_target(self):
        log.debug("Running target")

        # Setup and run target
        environment = [env for env in self.bh_instance.environments if env[0] == self.env_combobox.currentText()][0]
        source_target = self.bh_instance.targets[self.bh_instance.local_environment][0]
        if environment != source_target.environment:
            target = binharness.transport_target(source_target, environment[1])
        else:
            target = source_target

        if self.executor_combobox.currentText() == "None":
            executor = binharness.NullExecutor()

        process = executor.run_target(target)

        log.debug("Creating views")
        if self.executor_combobox.currentText() == "None":
            p_so = process.stdout
            stdout_view = StreamView(self.workspace, self.workspace.main_instance, "bottom", p_so, "stdout")
            log.debug("Adding stdout view")
            self.workspace.add_view(stdout_view)

        # Make stdout/stderr available to user

        log.debug("Target run")
        self.close()



from __future__ import annotations

import functools
import logging
from typing import TYPE_CHECKING

from angrmanagement.data.analysis_options import CFGForceScanMode
from angrmanagement.logic.threads import gui_thread_schedule_async

from .job import InstanceJob

if TYPE_CHECKING:
    from angrmanagement.data.instance import Instance
    from angrmanagement.logic.jobmanager import JobContext

_l = logging.getLogger(name=__name__)


class CFGGenerationJob(InstanceJob):
    """
    Job for generating the Control Flow Graph.
    """

    DEFAULT_CFG_ARGS = {
        "normalize": True,  # this is what people naturally expect
    }

    def __init__(self, instance: Instance, on_finish=None, **kwargs) -> None:
        super().__init__("CFG generation", instance, on_finish=on_finish)

        # TODO: sanitize arguments

        # make a copy
        cfg_args = dict(kwargs)
        for key, val in self.DEFAULT_CFG_ARGS.items():
            if key not in cfg_args:
                cfg_args[key] = val

        self.cfg_args = cfg_args

        scanning_mode = self.cfg_args.pop("scanning_mode", None)
        if scanning_mode is not None:
            self.cfg_args["force_smart_scan"] = scanning_mode == CFGForceScanMode.SmartScan
            self.cfg_args["force_complete_scan"] = scanning_mode == CFGForceScanMode.CompleteScan

        self._cfb = None

    def run(self, ctx: JobContext):
        exclude_region_types = {"kernel", "tls"}
        # create a temporary CFB for displaying partially analyzed binary during CFG recovery
        temp_cfb = self.instance.project.analyses.CFB(
            exclude_region_types=exclude_region_types, on_object_added=self._on_cfb_object_added
        )
        self._cfb = temp_cfb

        cfg = self.instance.project.analyses.CFG(
            progress_callback=functools.partial(self._progress_callback, ctx),
            low_priority=True,
            cfb=temp_cfb,
            **self.cfg_args,
        )
        self._cfb = None
        # Build the real one
        cfb = self.instance.project.analyses.CFB(kb=cfg.kb, exclude_region_types=exclude_region_types)

        return cfg.model, cfb

    def __repr__(self) -> str:
        return "Generating CFG"

    #
    # Private methods
    #

    def _progress_callback(self, ctx: JobContext, percentage, text: str | None = None, cfg=None) -> None:
        ctx.set_progress(percentage, text)

        if cfg is not None:
            # Peek into the CFG
            gui_thread_schedule_async(
                self._refresh,
                args=(
                    cfg,
                    self._cfb,
                ),
            )

    def _refresh(self, cfg, cfb) -> None:
        # do not signal events. that will happen on a timer to not overwhelm the renderer
        # instance will exist because _run must be used first
        self.instance.cfg = cfg
        self.instance.cfb = cfb

    def _on_cfb_object_added(self, addr: int, obj) -> None:
        self.instance.cfb.am_event(object_added=(addr, obj))



from __future__ import annotations

from typing import TYPE_CHECKING

from angr.knowledge_plugins.key_definitions.atoms import Atom, MemoryLocation, Register, SpOffset
from PySide6.QtCore import QRectF, Qt
from PySide6.QtGui import QColor, QPen
from PySide6.QtWidgets import QGraphicsSimpleTextItem

from angrmanagement.config import Conf
from angrmanagement.utils import get_string_for_display, locate_function

from .qgraph_object import QCachedGraphicsItem

if TYPE_CHECKING:
    import PySide6.QtWidgets
    from angr.knowledge_plugins.key_definitions.definition import Definition

    from angrmanagement.ui.views.dep_view import DependencyView


class QDepGraphBlock(QCachedGraphicsItem):
    HORIZONTAL_PADDING = 20
    VERTICAL_PADDING = 20
    LINE_MARGIN = 3

    def __init__(
        self,
        is_selected: bool,
        dep_view: DependencyView,
        definition: Definition = None,
        atom: Atom = None,
        addr: int = None,
    ) -> None:
        super().__init__()

        self._dep_view = dep_view
        self._instance = self._dep_view.instance
        self._config = Conf

        self.selected = is_selected

        # widgets
        self._definition_str: str = None
        self._definition_item: QGraphicsSimpleTextItem = None
        self._instruction_str: str = None
        self._instruction_item: QGraphicsSimpleTextItem = None
        self._function_str: str = None
        self._function_item: QGraphicsSimpleTextItem = None
        self._text: str | None = None
        self._text_item: QGraphicsSimpleTextItem | None = None

        self.definition = definition
        self.atom = atom
        self.addr = addr

        self._init_widgets()
        self._update_size()

        self.setAcceptHoverEvents(True)

    def _init_widgets(self) -> None:
        # definition
        self._definition_str = ""
        atom = self.definition.atom if self.definition is not None else self.atom
        addr_str = "unknown address" if self.addr is None else f"{self.addr:#x}"
        if isinstance(atom, Register):
            # convert it to a register name
            arch = self._instance.project.arch
            register_name = arch.translate_register_name(atom.reg_offset, size=atom.size)
            self._definition_str = f"Register {register_name} @ {addr_str}"
        elif isinstance(atom, MemoryLocation):
            if isinstance(atom.addr, SpOffset):
                self._definition_str = f"Stack sp{atom.addr.offset:+#x} @ {addr_str}"
            elif isinstance(atom.addr, int):
                self._definition_str = f"Memory {atom.addr:#x} @ {addr_str}"

        if not self._definition_str:
            # fallback
            self._definition_str = repr(self.definition.atom)

        # function and instruction text

        if self.addr is None:
            self._function_str = "Unknown"
            self._instruction_str = "Unknown"
        else:
            # function string
            the_func = locate_function(self._instance, self.addr)
            if the_func is None:
                # is it a SimProcedure?
                if self._instance.project.is_hooked(self.addr):
                    hooker = self._instance.project.hooked_by(self.addr)
                    self._function_str = "SimProcedure " + hooker.__class__.__name__.split(".")[-1]
                else:
                    self._function_str = "Unknown"
            else:
                offset = self.addr - the_func.addr
                if not the_func.name:
                    self._function_str = f"{the_func.addr:#x}{offset:+x}"
                else:
                    self._function_str = f"{the_func.name}{offset:+x}"
            # instruction
            self._instruction_str = f"{self._function_str}:  {self._instance.get_instruction_text_at(self.addr)}"
            # text
            self._text = get_string_for_display(
                self._instance.cfg,
                self.addr,
                self._instance.project,
                max_size=60,
            )

        x = self.HORIZONTAL_PADDING
        y = self.VERTICAL_PADDING

        # definition
        self._definition_item = QGraphicsSimpleTextItem(self._definition_str, self)
        self._definition_item.setBrush(Qt.GlobalColor.darkBlue)
        self._definition_item.setFont(Conf.symexec_font)
        self._definition_item.setPos(x, y)

        y += self._definition_item.boundingRect().height() + self.LINE_MARGIN

        # instruction
        self._instruction_item = QGraphicsSimpleTextItem(self._instruction_str, self)
        self._instruction_item.setBrush(Qt.GlobalColor.black)
        self._instruction_item.setFont(Conf.symexec_font)
        self._instruction_item.setPos(x, y)

        x += self._instruction_item.boundingRect().width()

        # text
        if self._text:
            x += 10
            self._text_item = QGraphicsSimpleTextItem(self._text, self)
            self._text_item.setFont(Conf.symexec_font)
            self._text_item.setBrush(Qt.GlobalColor.gray)
            self._text_item.setPos(x, y)

        # y += self._instruction_item.boundingRect().height()
        # x = self.HORIZONTAL_PADDING

    def refresh(self) -> None:
        self._update_size()

    #
    # Event handlers
    #

    def mousePressEvent(self, event) -> None:  # pylint: disable=no-self-use
        super().mousePressEvent(event)

    def mouseReleaseEvent(self, event) -> None:
        # _l.debug('QStateBlock received mouse release event')
        if event.button() == Qt.MouseButton.LeftButton:
            self.selected = not self.selected
            self._dep_view.redraw_graph()
            event.accept()

        super().mouseReleaseEvent(event)

    def mouseDoubleClickEvent(self, event) -> None:
        # _l.debug('QStateBlock received mouse double click event')
        if event.button() == Qt.MouseButton.LeftButton:
            self._workspace.viz(self.addr)
            event.accept()

        super().mouseDoubleClickEvent(event)

    def hoverEnterEvent(self, event: PySide6.QtWidgets.QGraphicsSceneHoverEvent) -> None:
        self._dep_view.hover_enter_block(self)

    def hoverLeaveEvent(self, event: PySide6.QtWidgets.QGraphicsSceneHoverEvent) -> None:
        self._dep_view.hover_leave_block()

    def paint(self, painter, option, widget) -> None:  # pylint: disable=unused-argument
        """
        Paint a state block on the scene.

        :param painter:
        :return: None
        """

        painter.setFont(Conf.symexec_font)
        normal_background = QColor(0xFA, 0xFA, 0xFA)
        selected_background = QColor(0xCC, 0xCC, 0xCC)

        # The node background
        if self.selected:
            painter.setBrush(selected_background)
        else:
            painter.setBrush(normal_background)
        painter.setPen(QPen(QColor(0xF0, 0xF0, 0xF0), 1.5))
        painter.drawRect(0, 0, self.width, self.height)

    def _boundingRect(self):
        return QRectF(0, 0, self._width, self._height)

    #
    # Private methods
    #

    def _update_size(self) -> None:
        width_candidates = [
            # definition string
            self._definition_item.boundingRect().width(),
            # instruction & text
            self.HORIZONTAL_PADDING * 2
            + self._instruction_item.boundingRect().width()
            + ((10 + self._text_item.boundingRect().width()) if self._text_item is not None else 0),
        ]

        self._width = max(width_candidates)
        self._height = (
            self.VERTICAL_PADDING * 2 + (self.LINE_MARGIN + self._definition_item.boundingRect().height()) * 2
        )

        self._width = max(100, self._width)
        self._height = max(50, self._height)

        self.recalculate_size()



from __future__ import annotations

import functools
import logging

import rpyc

from angrmanagement.logic import GlobalInfo
from angrmanagement.logic.threads import gui_thread_schedule_async

_l = logging.getLogger(name=__name__)


def requires_daemon_conn(f):
    @functools.wraps(f)
    def with_daemon_conn(self, *args, **kwargs):
        if self.conn is None:
            return None
        return f(self, *args, **kwargs)

    return with_daemon_conn


class ClientService(rpyc.Service):
    @property
    def instance(self):
        return GlobalInfo.main_window.workspace.main_instance

    @property
    def workspace(self):
        return GlobalInfo.main_window.workspace

    def exposed_jumpto(self, addr: int, symbol) -> None:
        if self.workspace is not None:
            gui_thread_schedule_async(GlobalInfo.main_window.bring_to_front)
            if addr is not None:
                gui_thread_schedule_async(self.workspace.jump_to, args=(addr,))
            elif symbol is not None:
                # TODO: Support it
                gui_thread_schedule_async(self.workspace.jump_to, args=(symbol,))

    def exposed_commentat(self, addr: int, comment: str) -> None:
        if self.workspace is not None and addr is not None:
            gui_thread_schedule_async(GlobalInfo.main_window.bring_to_front)
            gui_thread_schedule_async(self.workspace.set_comment(addr, comment))

    def exposed_custom_binary_aware_action(self, action, kwargs) -> None:  # pylint: disable=no-self-use
        kwargs_copy = dict(kwargs.items())  # copy it to local
        DaemonClient.invoke(action, kwargs_copy)


class DaemonClientCls:
    """
    Implements logic that the client needs to talk to the daemon service.
    """

    def __init__(self, custom_handlers=None) -> None:
        self.custom_handlers = {} if custom_handlers is None else custom_handlers

    def register_handler(self, action: str, handler) -> None:
        self.custom_handlers[action] = handler

    def invoke(self, action, kwargs) -> None:
        if action not in self.custom_handlers:
            _l.critical("Unregistered URL action %r", action)
            return
        self.custom_handlers[action](kwargs)

    @property
    def conn(self):
        return GlobalInfo.daemon_conn

    @requires_daemon_conn
    def register_binary(self, binary_name: str, target_id: str) -> None:
        self.conn.root.register_binary(binary_name, target_id)

    @requires_daemon_conn
    def exit(self) -> None:
        self.conn.exit()


DaemonClient = DaemonClientCls()



from __future__ import annotations

from PySide6.QtWidgets import QDialog, QDialogButtonBox, QHBoxLayout, QLabel, QVBoxLayout

from angrmanagement.ui.widgets import QAddressInput


class JumpTo(QDialog):
    """
    Dialog to jump to an address.
    """

    def __init__(self, disasm_view, parent=None) -> None:
        super().__init__(parent)

        # initialization
        self._disasm_view = disasm_view

        self._address_box = None
        self._status_label = None
        self._ok_button = None

        self.setWindowTitle("Jump to address")

        self.main_layout = QVBoxLayout()

        self._init_widgets()

        self.setLayout(self.main_layout)

        self.show()

    #
    # Private methods
    #

    def _init_widgets(self) -> None:
        # address label

        address_label = QLabel(self)
        address_label.setText("Address")

        address = QAddressInput(self._on_address_changed, self._disasm_view.instance, parent=self)
        self._address_box = address

        address_layout = QHBoxLayout()
        address_layout.addWidget(address_label)
        address_layout.addWidget(address)
        self.main_layout.addLayout(address_layout)

        # status label
        status_label = QLabel(self)
        self.main_layout.addWidget(status_label)
        self._status_label = status_label

        # buttons
        buttons = QDialogButtonBox(parent=self)
        buttons.setStandardButtons(QDialogButtonBox.StandardButton.Cancel | QDialogButtonBox.StandardButton.Ok)
        buttons.accepted.connect(self._on_ok_clicked)
        buttons.rejected.connect(self.close)
        self._ok_button = buttons.button(QDialogButtonBox.StandardButton.Ok)
        self._ok_button.setEnabled(False)
        self.main_layout.addWidget(buttons)

    #
    # Event handlers
    #

    def _on_address_changed(self, new_text) -> None:  # pylint: disable=unused-argument
        if self._address_box.target is None:
            # the address is invalid

            self._status_label.setText("Invalid")
            self._status_label.setProperty("class", "status_invalid")
            self._ok_button.setEnabled(False)
        else:
            self._status_label.setText("Valid")
            self._status_label.setProperty("class", "status_valid")
            self._ok_button.setEnabled(True)

        self._status_label.style().unpolish(self._status_label)
        self._status_label.style().polish(self._status_label)

    def _on_ok_clicked(self) -> None:
        addr = self._address_box.target
        if addr is not None:
            r = self._disasm_view.jump_to(addr)
            if r:
                self.close()



from __future__ import annotations

import base64
import socket
from typing import TYPE_CHECKING

import claripy
from angr.storage.file import SimPacketsStream
from PySide6.QtCore import QAbstractItemModel, QModelIndex, QSize, Qt
from PySide6.QtGui import QColor, QContextMenuEvent, QIntValidator
from PySide6.QtWidgets import (
    QComboBox,
    QDialog,
    QHBoxLayout,
    QLineEdit,
    QMenu,
    QPushButton,
    QStyledItemDelegate,
    QTextEdit,
    QTreeView,
    QVBoxLayout,
)

if TYPE_CHECKING:
    from angrmanagement.data.instance import Instance

_socket_families_wanted = ["AF_INET", "AF_INET6", "AF_UNIX", "AF_CAN", "AF_PACKET", "AF_RDS"]

socket_family = {s: getattr(socket, s) for s in _socket_families_wanted if s in dir(socket)}
socket_type = {"SOCK_STREAM": socket.SOCK_STREAM, "SOCK_DGRAM": socket.SOCK_DGRAM, "SOCK_RAW": socket.SOCK_RAW}


class SocketItem:  # pylint: disable=no-self-use, unused-argument
    """
    Socket Item for SocketModel
    """

    def __init__(self, ident=None, parent=None, node_type=None) -> None:
        self.parentItem = parent
        self.children = []
        self.ident = ident
        self.recv_pkg = None
        self.node_type = node_type

    def appendChild(self, item) -> None:
        self.children.append(item)

    def child(self, row):
        return self.children[row]

    def childCount(self):
        return len(self.children)

    def columnCount(self) -> int:
        return 1

    def data(self, column):
        if column == 0:
            if self.node_type in ("Socket", "Accepted"):
                return self.ident
            else:  # self.Node_Type == "Package"
                return self.recv_pkg
        else:
            return None

    def setData(self, column, data) -> bool:
        if column == 0:
            if self.node_type in ("Socket", "Accepted"):
                self.ident = data
            else:  # self.Node_Type == "Package"
                self.recv_pkg = data
        return True

    def parent(self):
        return self.parentItem

    def row(self):
        if self.parentItem:
            return self.parentItem.children.index(self)

        return 0


class SimPackagePersistentEditor(QStyledItemDelegate):
    """
    Table Cell Editor
    """

    # pylint: disable=no-self-use, unused-argument
    def __init__(self, parent=None, instance: Instance | None = None) -> None:
        super().__init__(parent)
        self.instance = instance

    def createEditor(self, widget, option, index):
        editor = QTextEdit(widget)
        return editor

    def setEditorData(self, editor, index) -> None:
        value = index.model().data(index, role=Qt.ItemDataRole.DisplayRole)
        if value:
            editor.setText(str(value))

    def setModelData(self, editor, model, index) -> None:
        model.setData(index, editor.toPlainText(), role=Qt.ItemDataRole.DisplayRole)

    def updateEditorGeometry(self, editor, option, index) -> None:
        editor.setGeometry(option.rect)


class SocketModel(QAbstractItemModel):
    """
    Abstract Model with SocketView
    """

    # pylint: disable=no-self-use, unused-argument
    def __init__(self, data=None, parent=None) -> None:
        super().__init__(parent)
        self.rootItem = SocketItem()

    def columnCount(self, parent=None):
        if parent is None:
            parent = QModelIndex()
        if parent.isValid():
            return parent.internalPointer().columnCount()
        else:
            return self.rootItem.columnCount()

    def data(self, index, role=None):
        if not index.isValid():
            return None

        if role == Qt.SizeHintRole:
            if index.internalPointer().node_type == "Package":
                return QSize(300, 100)
            else:
                return QSize(300, 20)
        if role == Qt.BackgroundRole:
            if index.internalPointer().node_type == "Package":
                return QColor(226, 237, 253)
            else:
                return Qt.white

        if role != Qt.DisplayRole:
            return None

        item = index.internalPointer()

        return item.data(index.column())

    def setData(self, index, data, role=None):
        if not index.isValid():
            return None

        if role != Qt.DisplayRole:
            return None

        item = index.internalPointer()

        return item.setData(index.column(), data)

    def flags(self, index):
        if not index.isValid():
            return Qt.NoItemFlags
        if index.internalPointer().node_type == "Package":
            return Qt.ItemIsEnabled | Qt.ItemIsSelectable | Qt.ItemIsEditable
        return Qt.ItemIsEnabled | Qt.ItemIsSelectable

    def headerData(self, section, orientation, role=None):
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            return "Socket"

        return None

    def index(self, row, column, parent=None):
        if parent is None:
            parent = QModelIndex()

        if not self.hasIndex(row, column, parent):
            return QModelIndex()

        parentItem = self.rootItem if not parent.isValid() else parent.internalPointer()

        childItem = parentItem.child(row)
        if childItem:
            return self.createIndex(row, column, childItem)
        else:
            return QModelIndex()

    def rowCount(self, parent=None):
        if parent is None:
            parent = QModelIndex()

        if parent.column() > 0:
            return 0

        parentItem = self.rootItem if not parent.isValid() else parent.internalPointer()

        return parentItem.childCount()

    def parent(self, index):
        if not index.isValid():
            return QModelIndex()

        childItem = index.internalPointer()
        parentItem = childItem.parent()

        if parentItem == self.rootItem:
            return QModelIndex()

        return self.createIndex(parentItem.row(), 0, parentItem)

    def add_item(self, ident, parent=None, node_type=None) -> None:
        if parent is None:
            parent = QModelIndex()
        self.beginInsertRows(parent, self.rowCount(parent), self.rowCount(parent))
        parentItem = self.rootItem if not parent.isValid() else parent.internalPointer()
        item = SocketItem(ident, parent=parentItem, node_type=node_type)
        parentItem.appendChild(item)
        self.endInsertRows()

    def del_item(self, item) -> None:
        parent = item.parent()
        parentItem = self.rootItem if not parent.isValid() else parent.internalPointer()
        self.beginRemoveRows(parent, item.row(), item.row())
        del parentItem.children[item.row()]
        self.endRemoveRows()

    def get_data(self):
        ret = {}
        for socket_info in self.rootItem.children:
            tmp = []
            for child in socket_info.children:
                if child.node_type == "Package":
                    tmp.append(child.recv_pkg)
                elif child.node_type == "Accepted":
                    ret[child.ident] = [p.recv_pkg for p in child.children]
            if tmp:
                ret[socket_info.ident] = tmp
        return ret

    def convert_ident(self, ident):
        return ("socket", socket_family[ident[0]], socket_type[ident[1]], 0, int(ident[2]))

    def convert(self):
        data = self.get_data()
        ret = {}
        for k, v in data.items():
            ident = k.split(", ")
            if ident[0] == "Accept":
                ident = ("accept", self.convert_ident(ident[1:]), int(ident[4]))
            else:
                ident = self.convert_ident(ident)
            content = [claripy.BVV(base64.b64decode(x)) for x in v]
            ret[ident] = (SimPacketsStream("socket read", content=content), SimPacketsStream("socket write"))
        return ret


class SocketView(QTreeView):
    """
    Socket Config Tree View with SocketModel
    """

    def __init__(self) -> None:
        super().__init__()
        self.setTextElideMode(Qt.TextElideMode.ElideNone)

    def _action_accepted_socket(self) -> None:
        current = self.currentIndex()
        if current.parent().isValid():
            current = current.parent()
        ident = "Accept, " + current.internalPointer().ident + (", %d" % (current.internalPointer().childCount() + 1))
        self.model().add_item(ident, current, "Accepted")

    def _action_add_package(self) -> None:
        current = self.currentIndex()
        if current.internalPointer().node_type == "Package":
            current = current.parent()
        ident = "Package"
        self.model().add_item(ident, current, "Package")

    def _action_delete(self) -> None:
        current = self.currentIndex()
        self.model().del_item(current)

    def contextMenuEvent(self, event: QContextMenuEvent) -> None:
        menu = QMenu("", self)
        menu.addAction("Add an accepted socket", self._action_accepted_socket)
        menu.addAction("Add a recv packages", self._action_add_package)
        menu.addAction("Delete", self._action_delete)
        menu.exec_(event.globalPos())


class SocketConfig(QDialog):
    """
    Socket Config Dialog
    """

    family = list(socket_family.keys())
    typ = list(socket_type.keys())

    def __init__(self, socket_config=None, instance: Instance | None = None, parent=None) -> None:
        super().__init__(parent)

        self.setWindowFlags(self.windowFlags() & ~Qt.WindowType.WindowContextHelpButtonHint)
        self._instance = instance
        self._editor = SimPackagePersistentEditor(instance=instance)
        self._parent = parent
        self.socket_config: SocketModel
        if socket_config:
            self.socket_config = socket_config
        else:
            self.socket_config = SocketModel()
        self._init_widgets()

    def _init_widgets(self) -> None:
        layout = QVBoxLayout()
        self._table = SocketView()
        self._table.setModel(self.socket_config)
        self._table.setItemDelegate(self._editor)

        layout.addWidget(self._table, 0)
        toolbox = QHBoxLayout()
        self._socket_family = QComboBox(self)
        self._socket_family.addItems(["Family"] + self.family)
        self._socket_type = QComboBox(self)
        self._socket_type.addItems(["Type"] + self.typ)
        self._socket_nonce = QLineEdit(self)
        self._socket_nonce.setValidator(QIntValidator(0, 9999, self))
        self._socket_add_button = QPushButton("Add new socket")

        def _add_new_socket() -> None:
            family = self._socket_family.currentIndex() - 1
            typ = self._socket_family.currentIndex() - 1
            nonce = self._socket_nonce.text()
            if family < 0 or typ < 0 or not nonce:
                return
            txt = ", ".join([self._socket_family.currentText(), self._socket_type.currentText(), nonce])
            self.socket_config.add_item(txt, node_type="Socket")

        self._socket_add_button.clicked.connect(_add_new_socket)

        toolbox.addWidget(self._socket_family, 0)
        toolbox.addWidget(self._socket_type, 1)
        toolbox.addWidget(self._socket_nonce, 2)
        toolbox.addWidget(self._socket_add_button, 3)
        layout.addLayout(toolbox, 1)
        self.setLayout(layout)

    def closeEvent(self, event) -> None:  # pylint: disable=unused-argument
        # print(self.socket_config.get_data())
        self.close()



from __future__ import annotations

from typing import TYPE_CHECKING

from pyqodeng.core.api import CodeEdit
from pyqodeng.core.modes import AutoIndentMode, CaretLineHighlighterMode, PygmentsSyntaxHighlighter
from PySide6.QtGui import QTextOption
from PySide6.QtWidgets import (
    QDialog,
    QDialogButtonBox,
    QGridLayout,
    QGroupBox,
    QLabel,
    QRadioButton,
    QScrollArea,
    QVBoxLayout,
    QWidget,
)

if TYPE_CHECKING:
    from angrmanagement.ui.workspace import Workspace


class HookDialog(QDialog):
    """
    Provide templetes of hook function.
    """

    def __init__(self, workspace: Workspace, addr: int | None = None, parent=None) -> None:
        super().__init__(parent)

        # initialization

        self.workspace = workspace
        self.state = None  # output
        self._addr = addr
        self.templates = {}
        self._add_templates(hex(addr))
        self._function_box = None
        self._ok_button = None
        self.setWindowTitle("Change hook")
        self.main_layout = QVBoxLayout()
        self._init_widgets()
        self.setLayout(self.main_layout)

    #
    # Private methods
    #

    def _add_templates(self, addr: int) -> None:
        self.templates[
            "base"
        ] = f"""\
@project.hook(addr={addr}, length=0)
def hook(state):
    ..."""

        self.templates[
            "assertion"
        ] = f"""\
@project.hook(addr={addr}, length=0)
def assertion(state):
    state.add_constraints(
        ...
    )"""

        self.templates[
            "disable unicorn"
        ] = f"""\
@project.hook(addr={addr}, length=0)
def disable_unicorn(state):
    state.options.discard("UNICORN")
    state.options.discard("UNICORN_HANDLE_TRANSMIT_SYSCALL")
    state.options.discard("UNICORN_SYM_REGS_SUPPORT")
    state.options.discard("UNICORN_TRACK_BBL_ADDRS")
    state.options.discard("UNICORN_TRACK_STACK_POINTERS")
"""

        self.templates[
            "enable unicorn"
        ] = f"""\
@project.hook(addr={addr}, length=0)
def enable_unicorn(state):
    state.options.add("UNICORN")
    state.options.add("UNICORN_HANDLE_TRANSMIT_SYSCALL")
    state.options.add("UNICORN_SYM_REGS_SUPPORT")
    state.options.add("UNICORN_TRACK_BBL_ADDRS")
    state.options.add("UNICORN_TRACK_STACK_POINTERS")
"""

    def update_function(self, template) -> None:
        self._function_box.setPlainText(template, mime_type="text/x-python", encoding="utf-8")

    def selected(self) -> None:
        btn = self.sender()
        if btn.isChecked():
            self.update_function(btn.template)

    def _init_widgets(self) -> None:
        layout = QGridLayout()
        row = 0

        # validation_failures = set()
        addr = hex(self._addr)
        address_label = QLabel(self)
        address_label.setText(f"Hook at address {addr}:")

        layout.addWidget(address_label, row, 0)
        row += 1

        options_container = QGroupBox(self)
        options = QVBoxLayout()

        for name, template in sorted(self.templates.items()):
            child = QRadioButton()
            child.setText(name)
            child.template = template
            child.toggled.connect(self.selected)
            options.addWidget(child)

        scroll_area = QScrollArea(self)
        scroll_area.setWidgetResizable(True)
        widget = QWidget()
        scroll_area.setWidget(widget)
        layout_scroll = QVBoxLayout(widget)

        header_label = QLabel(self)
        header_label.setText("Presets:")

        layout_scroll.addWidget(header_label)
        layout_scroll.addWidget(options_container)
        options_container.setLayout(options)
        layout.addWidget(scroll_area, row, 0)
        row += 1

        function_box = CodeEdit(self)
        function_box.use_spaces_instead_of_tabs = True
        function_box.tab_length = 4

        function_box.modes.append(CaretLineHighlighterMode())
        function_box.modes.append(PygmentsSyntaxHighlighter(function_box.document()))
        function_box.modes.append(AutoIndentMode())

        function_box.setWordWrapMode(QTextOption.WrapMode.WordWrap)
        self._function_box = function_box

        layout.addWidget(function_box, row, 0)
        row += 1

        self.main_layout.addLayout(layout)

        buttons = QDialogButtonBox(parent=self)
        buttons.setStandardButtons(QDialogButtonBox.StandardButton.Cancel | QDialogButtonBox.StandardButton.Ok)
        buttons.button(QDialogButtonBox.StandardButton.Ok).setText("Append to Console")

        def do_ok() -> None:
            code = function_box.toPlainText()
            self.workspace.append_code_to_console(code)
            self.close()

        buttons.accepted.connect(do_ok)
        buttons.rejected.connect(self.close)
        self.main_layout.addWidget(buttons)



from __future__ import annotations

import os

from PySide6.QtCore import Qt
from PySide6.QtGui import QFont, QIcon, QPixmap
from PySide6.QtWidgets import QDialog, QHBoxLayout, QLabel, QVBoxLayout

import angrmanagement
from angrmanagement.config import IMG_LOCATION


class LoadAboutDialog(QDialog):
    """
    Dialog that shows application version, credits, etc.
    """

    def __init__(self) -> None:
        super().__init__()
        self.setWindowFlags(Qt.WindowType.WindowTitleHint | Qt.WindowType.WindowCloseButtonHint)
        self.setWindowTitle("About")
        # mdiIcon
        angr_icon_location = os.path.join(IMG_LOCATION, "angr.png")
        self.setWindowIcon(QIcon(angr_icon_location))
        self._init_widgets()

    def _init_widgets(self) -> None:
        # icon
        icon_label = QLabel(self)
        icon_location = os.path.join(IMG_LOCATION, "angr-ds.png")
        angr_icon = QPixmap(icon_location)
        icon_label.setPixmap(angr_icon)
        # textbox
        angr_text = QLabel("angr")
        angr_text.setFont(QFont("Consolas", 24, weight=QFont.Weight.Bold))
        version_text_tup = "Version: " + angrmanagement.__version__
        version_text = QLabel(version_text_tup)
        version_text.setFont(QFont("Consolas", weight=QFont.Weight.Bold))
        version_text.setAlignment(Qt.AlignmentFlag.AlignCenter)
        version_text.setTextInteractionFlags(Qt.TextInteractionFlag.TextSelectableByMouse)
        credits_text = QLabel('<a href="http://angr.io/">Credits</a>')
        credits_text.setFont(QFont("Consolas", weight=QFont.Weight.Bold))
        credits_text.setTextFormat(Qt.TextFormat.RichText)
        credits_text.setTextInteractionFlags(Qt.TextInteractionFlag.TextBrowserInteraction)
        credits_text.setOpenExternalLinks(True)

        structure = QVBoxLayout()
        structure.addWidget(angr_text)
        structure.addWidget(version_text)
        structure.addWidget(credits_text)

        layout = QHBoxLayout()
        layout.addWidget(icon_label)
        layout.addLayout(structure)

        self.setLayout(layout)

        #
        # Event handlers
        #



from __future__ import annotations

import logging
import subprocess

from PySide6.QtWidgets import QInputDialog, QMessageBox

_l = logging.getLogger(__name__)


class LoadDockerPromptError(Exception):
    pass


class LoadDockerPrompt(QInputDialog):
    def __init__(self, parent=None) -> None:
        super().__init__(parent)

        self.setComboBoxItems(get_docker_images(self))
        self.setOption(QInputDialog.InputDialogOption.UsePlainTextEditForTextInput, True)
        self.setLabelText("Select a docker target")
        self.setComboBoxEditable(False)
        self.setModal(True)
        _l.info("Docker prompt succesfully loaded")


def get_docker_images(parent=None):
    # Try to list docker images, handle failure cases
    try:
        output = subprocess.check_output(["docker", "images", "--format", "{{.Repository}}:{{.Tag}}"])
    except FileNotFoundError as e:
        _l.error("Docker not found.")
        QMessageBox(parent).critical(None, "Docker not found", "Make sure docker is installed and in your path")
        raise LoadDockerPromptError from e
    except subprocess.CalledProcessError as e:
        _l.error("Docker images failed: Make sure you are have privileges.")
        QMessageBox(parent).critical(
            None, "Docker Error", "User not able to access docker. Are you in the docker group?"
        )
        raise LoadDockerPromptError from e

    items = output.decode("utf-8").split("\n")
    return [i for i in items if i and "<none>" not in i]



# pylint:disable=missing-class-docstring
from __future__ import annotations

import logging
import os
import random
from bisect import bisect_left
from collections import defaultdict
from typing import TYPE_CHECKING, Any

from angr.errors import SimEngineError
from PySide6.QtGui import QColor

if TYPE_CHECKING:
    from angrmanagement.ui.workspace import Workspace

log = logging.getLogger(name=__name__)


class TraceFunc:
    __slots__ = (
        "bbl_addr",
        "func_name",
        "func",
    )

    def __init__(self, bbl_addr=None, func_name: str | None = None, func=None) -> None:
        self.bbl_addr = bbl_addr
        self.func_name = func_name
        self.func = func


class ObjectAndBase:
    __slots__ = (
        "obj_name",
        "base_addr",
    )

    def __init__(self, obj_name: str, base_addr: int) -> None:
        self.obj_name = obj_name
        self.base_addr = base_addr

    def __lt__(self, other):
        if isinstance(other, ObjectAndBase):
            return self.base_addr < other.base_addr
        elif isinstance(other, int):
            return self.base_addr < other
        raise TypeError(f"Unsupported type {type(other)}")


class TraceStatistics:
    BBL_FILL_COLOR = QColor(0, 0xF0, 0xF0, 0xF)
    BBL_BORDER_COLOR = QColor(0, 0xF0, 0xF0)
    BBL_EMPTY_COLOR = QColor("white")

    def __init__(self, workspace: Workspace, trace, baddr) -> None:
        self.workspace = workspace
        self.trace: dict[str, Any] = trace
        self.bbl_addrs = trace["bb_addrs"]
        self.syscalls = trace["syscalls"]
        self.id = trace["id"]
        self.created_at = trace["created_at"]
        self.input_id = trace["input_id"]
        self.complete = trace["complete"]
        self.mapping: list[ObjectAndBase] | None = None
        if "map" in trace:
            map_dict: dict[str, int] = trace["map"]
            self.mapping = [ObjectAndBase(name, base_addr) for name, base_addr in map_dict.items()]
            self.mapping = sorted(self.mapping, key=lambda o: o.base_addr)  # sort it based on base addresses
        self.trace_func: list[TraceFunc] = []
        self.func_addr_in_trace: set[int] = set()
        self._func_color = {}
        self.count = None
        self._mark_color = {}
        self._positions = defaultdict(list)
        self.mapped_trace = []

        self.project = self.workspace.main_instance.project
        if self.project.am_none:
            self.project_baddr = None
        else:
            # only used if self.mapping is not available
            self.project_baddr = self.project.loader.main_object.mapped_base
        self.runtime_baddr = baddr  # this will not be used if self.mapping is available

        self._cached_object_project_base_addrs: dict[str, int] = {}

        self._statistics(self.bbl_addrs)

    def find_object_base_in_project(self, object_name: str) -> int | None:
        """
        Find the base address of an object in angr project. Returns None if the project is not mapped. Results are
        cached in self._cached_object_project_base_addrs.

        :param object_name: Name of the object to look for.
        :return:            The base address of the loaded object in angr Project.
        """

        try:
            return self._cached_object_project_base_addrs[object_name]
        except KeyError:
            pass

        base_addr = None
        base_obj_name = os.path.basename(object_name)
        for obj in self.project.loader.all_objects:
            if not hasattr(obj, "binary"):
                continue
            if obj.binary and os.path.basename(obj.binary) == base_obj_name:
                # found it!
                # we assume binary names are unique. if they are not, someone should the logic here.
                base_addr = obj.mapped_base
                break

        if base_addr is None:
            log.warning(
                "Cannot find object %s in angr project. Maybe it has not been loaded. Exclude it from the trace.",
                object_name,
            )
        self._cached_object_project_base_addrs[object_name] = base_addr
        return base_addr

    def get_func_color(self, func_name: str):
        if func_name in self._func_color:
            return self._func_color[func_name]
        else:
            color = self._random_color()
            self._func_color[func_name] = color
        return color

    def set_mark_color(self, p, color) -> None:
        self._mark_color[p] = color

    def get_mark_color(self, addr: int, i):
        try:
            mark_index = self._get_position(addr, i)
            mark_color = self._mark_color[mark_index]
        except (IndexError, KeyError) as e:
            log.error(e)
            return self.BBL_EMPTY_COLOR
        return mark_color

    def get_positions(self, addr: int):
        return self._positions[addr]

    def get_count(self, ins):
        return len(self._positions[ins])

    def get_bbl_from_position(self, position):
        return self.trace_func[position].bbl_addr

    def get_func_name_from_position(self, position):
        return self.trace_func[position].func_name

    def get_func_from_position(self, position):
        return self.trace_func[position].func

    def _apply_trace_offset(self, addr: int) -> int | None:
        if self.mapping is not None and self.mapping:
            # find the base address that this address belongs to
            idx = bisect_left(self.mapping, addr)
            obj = None
            if 0 <= idx < len(self.mapping):
                # check if addr == object.base
                obj = self.mapping[idx]
                if addr == obj.base_addr:
                    # yes
                    pass
                elif idx > 0:
                    obj = self.mapping[idx - 1]
            else:  # idx == len(self.mapping)
                obj = self.mapping[idx - 1]

            if obj is not None:
                project_base_addr = self.find_object_base_in_project(obj.obj_name)
                if project_base_addr is not None:
                    return addr + (project_base_addr - obj.base_addr)
                else:
                    # not found - the object is probably not loaded in angr? ignore it
                    return None

        # fall back
        if self.project_baddr is not None:
            offset = self.project_baddr - self.runtime_baddr
            return addr + offset
        else:
            # this object is probably created before an angr project is created. just give up.
            return None

    def _statistics(self, trace_addrs) -> None:
        """
        :param trace: basic block address list
        """
        self.mapped_trace = []
        for addr in trace_addrs:
            converted_addr = self._apply_trace_offset(addr)
            if converted_addr is not None:
                self.mapped_trace.append(converted_addr)

        bbls = filter(self._get_bbl, self.mapped_trace)
        functions = self.workspace.main_instance.project.kb.functions

        for p, bbl_addr in enumerate(bbls):
            node = self.workspace.main_instance.cfg.get_any_node(bbl_addr)
            # if node is None:  # try again without assuming node is start of a basic block
            #     node = self.workspace.instance.cfg.get_any_node(bbl_addr, anyaddr=True)

            func = None
            if node is not None:
                if node.instruction_addrs is not None:
                    instr_addrs = node.instruction_addrs
                else:
                    # relift
                    block = self.workspace.main_instance.project.factory.block(bbl_addr)
                    instr_addrs = block.instruction_addrs
                for addr in instr_addrs:
                    self._positions[addr].append(p)

                func_addr = node.function_address
                if func_addr is not None and functions.contains_addr(func_addr):
                    func = functions.get_by_addr(func_addr)
                    func_name = func.demangled_name
                else:
                    func_name = "Unknown"
                self.func_addr_in_trace.add(func_addr)
            else:
                # Node is not found in the CFG. It's possible that the library is not loaded
                func_name = hex(bbl_addr)  # default to using bbl_addr as name if none is not found
                # log.warning("Node at %x is None, using bbl_addr as function name", bbl_addr)
            self.trace_func.append(TraceFunc(bbl_addr, func_name, func))

            if p % 5000 == 0:
                print("... trace loading progress: %.02f%%" % (p * 100 / len(self.mapped_trace)))

        print("Trace is loaded.")
        self.count = len(self.trace_func)

    def _get_bbl(self, addr: int):
        try:
            return self.workspace.main_instance.project.factory.block(addr)
        except SimEngineError:
            return None

    def _func_addr(self, a):
        return self.workspace.main_instance.cfg.get_any_node(a).function_address

    def _func_name(self, a):
        return self.workspace.main_instance.project.kb.functions[self._func_addr(a)].demangled_name

    @staticmethod
    def _random_color():
        r = random.randint(0, 255)
        g = random.randint(0, 255)
        b = random.randint(0, 255)
        return QColor(r, g, b)

    def _get_position(self, addr: int, i):
        return self._positions[addr][i]



from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING

from PySide6.QtCore import Qt
from PySide6.QtWidgets import QMessageBox

if TYPE_CHECKING:
    from angrmanagement.data.instance import Instance
    from angrmanagement.ui.views import DisassemblyView
    from angrmanagement.ui.widgets.qblock import QBlock
    from angrmanagement.ui.widgets.qoperand import QOperand


class DisassemblyLevel(Enum):
    MachineCode = 0
    LifterIR = 1
    AIL = 2


class QDisassemblyBaseControl:
    """
    The base control class of QLinearViewer and QDisassemblyGraph. Implements or declares common shorthands and methods.
    """

    def __init__(self, instance: Instance, disasm_view, base_cls) -> None:
        self.instance = instance
        self.disasm_view: DisassemblyView = disasm_view
        self._base_cls = base_cls
        self._insaddr_to_block = {}
        self._disassembly_level = disasm_view.disassembly_level

    @property
    def infodock(self):
        return self.disasm_view.infodock

    def refresh(self):
        """
        Recalculate sizes and positions of subitems, and trigger a refresh on every subitem (in order to reload text,
        etc.)

        :return:    None
        """
        raise NotImplementedError

    def reload(self):
        raise NotImplementedError

    def show_instruction(
        self,
        insn_addr,
        insn_pos=None,
        centering: bool = False,
        use_block_pos: bool = False,
        use_animation: bool = False,
    ):
        raise NotImplementedError

    #
    # Public methods
    #

    def get_selected_operand_info(self) -> tuple[QBlock, int, QOperand] | None:
        if not self.infodock.selected_operands:
            return None

        # get the first operand
        ins_addr, operand_idx = next(iter(self.infodock.selected_operands))
        block = self._insaddr_to_block.get(ins_addr, None)
        if block is not None:
            operand = block.addr_to_insns[ins_addr].get_operand(operand_idx)
            return block, ins_addr, operand

        return None

    def set_disassembly_level(self, level: DisassemblyLevel) -> None:
        self._disassembly_level = level
        self.reload()

    #
    # Event handlers
    #

    def keyPressEvent(self, event) -> None:
        key = event.key()

        if key == Qt.Key.Key_N:
            # rename a label
            self.disasm_view.popup_rename_label_dialog()
            return
        elif key == Qt.Key.Key_X:
            # XRef
            # try to get a selected operand
            r = self.get_selected_operand_info()
            if r is not None:
                # xref to an operand
                _, ins_addr, operand = r
                self.disasm_view.parse_operand_and_popup_xref_dialog(ins_addr, operand)
                return

            # try to get a selected label
            if len(self.infodock.selected_labels) == 1:
                lbl_addr = next(iter(self.infodock.selected_labels))
                self.disasm_view.popup_xref_dialog(addr=lbl_addr, dst_addr=lbl_addr)
                return

            # try to get a selected variable
            if len(self.infodock.selected_variables) == 1:
                variable = next(iter(self.infodock.selected_variables))
                self.disasm_view.popup_xref_dialog(addr=0, variable=variable)
                return

            # message the user
            QMessageBox.critical(
                None,
                "Invalid selection for XRefs",
                "You must select an operand, a label, a variable, or a function header before requesting XRefs.",
            )

            return

        self._base_cls.keyPressEvent(self, event)
